/*
 * $Id$
 *
 * Copyright 2002 John Pritchard, All Rights Reserved.
 *
 * Licensed under the "Common Public License 1.0".
 *  http://www.yajre.org/license.txt
 */
/*
 * LICENSE.txt,v 1.1 2001/12/12 19:33:28 werken 
 *
 * Copyright 2001 (C) The Werken Company. All Rights Reserved.
 * 
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 * 
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 * 
 * 3. The name "Drools" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of The Werken Company.  For written permission,
 *    please contact bob@werken.com.
 * 
 * 4. Products derived from this Software may not be called "Drools"
 *    nor may "Drools" appear in their names without prior written
 *    permission of The Werken Company. Drools is a registered
 *    trademark of The Werken Company.
 * 
 * 5. Due credit should be given to the Drools Project
 *    (http://drools.org/).
 * 
 * THIS SOFTWARE IS PROVIDED BY THE WERKEN COMPANY AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE WERKEN COMPANY OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */

package org.yajre;

import org.yajre.reteoo.RootNode;
import org.yajre.reteoo.Builder;
import org.yajre.reteoo.ReteConstructionException;

import org.yajre.spi.ObjectType;
import org.yajre.spi.Rule;
import org.yajre.spi.RuleSet;
import org.yajre.spi.Declaration;

import java.util.* ;

/**
 * Collection of {@link Rule}s.
 *
 * @author <a href="bob@werken.com">bob@werken.com</a>
 * @author John Pritchard (john@syntelos.org)
 */
public class RuleBase {

    /**
     * The root Rete-OO node of this <code>RuleBase</code>.
     */
    private final RootNode rootnode = new RootNode();

    /**
     * Transition Rules: whether sessions on this rule base are
     * holding actions for explicit post- assertions' firing.
     */
    private boolean transition = false;

    /**
     * Rule set names
     */
    private final List rulesets = new ArrayList();

    /**
     * Optional application attributes of this rule base available to
     * I/O integration implementors.
     */
    private Map attributes = new HashMap();

    /** 
     * A new rule base not using `transition rules'.
     */
    public RuleBase(){
	super();
    }

    /** 
     * @param transition If true, use `transition rules'
     * holding actions to be fired for explicit firing after
     * assertions.
     */
    public RuleBase(boolean transition){
	super();

	this.transition = transition;
    }

    /**
     * Get the map for the attributes of this rule base.  (Used by
     * integration components for application purposes.)
     */
    public final Map attributes(){ return this.attributes;}

    /**
     * Assign the map for the attributes of this rule base.  (Used by
     * integration components for application purposes.)
     */
    public final void attributes( Map reset){

	if ( null != reset)
	    this.attributes = reset;
    }

    /**
     * Lookup an attribute of this rule base.  (Used by integration
     * components for application purposes.)
     */
    public final Object getAttribute( Object key){

	return this.attributes.get(key);
    }

    /**
     * Define an attribute of this rule base.  (Used by integration
     * components for application purposes.)
     */
    public final void setAttribute( Object key, Object value){

	this.attributes.put(key,value);
    }

    /**
     * Delete an attribute of this rule base.  (Used by integration
     * components for application purposes.)
     */
    public final void removeAttribute( Object key){

	this.attributes.remove(key);
    }

    /**
     * Given an (user) type name or parameter identifier, return its
     * declaration if known.
     *
     * @param id Fact identity name.
     *
     * @returns Null if not found.
     */
    public final Declaration identity ( String id){

	return this.rootnode.identity(id);
    }

    /**
     * Given an (user) type name or parameter identifier, return or
     * create its declaration.
     * 
     * @param type Fact identity type.
     * 
     * @param id Fact identity name.
     *
     * @returns Declaration
     */
    public final Declaration identity ( ObjectType type, String id){

	return this.rootnode.identity(type,id);
    }

    /** 
     * Add a <code>RuleSet</code> of <code>Rules</code> to this
     * <code>RuleBase</code>.
     *
     * <p> A <code>RuleSet</code> may be added to multiple
     * <code>RuleBases</code>.  Any changes to a <code>RuleSet</code>
     * or its component <code>Rule</code> once it has been added are
     * ignored.
     *
     * @param ruleSet The <code>RuleSet</code> to add.
     *
     * @throws ReteConstructionException If a member rule does not
     * allow for complete and correct integration into the underlying
     * Rete network.
     */
    public final void add(RuleSet ruleSet) throws ReteConstructionException {

        Iterator ruleIter = ruleSet.getRuleIterator();

	if ( ruleIter.hasNext()){

	    rulesets.add(ruleSet.getName());

	    Rule rule;

	    do {
		rule = (Rule)ruleIter.next();

		Builder.AddRule(this.rootnode, rule);
	    }
	    while ( ruleIter.hasNext());

	    return;
	}
	else
	    return;
    }

    /**
     * Add a <code>Rule</code> to this <code>RuleBase</code>.
     *
     * <p> A <code>Rule</code> may be added to multiple
     * <code>RuleBases</code>.  Any changes to a <code>Rule</code>
     * once it has been added are ignored.
     *
     * @param rule The rule to add.
     *
     * @throws ReteConstructionException If the rule does not allow
     * for complete and correct integration into the underlying Rete
     * network.
     */
    public final void add(Rule rule) throws ReteConstructionException {

        Builder.AddRule( this.rootnode, rule );
    }

    /**
     * Create a {@link WorkingMemory} session for this
     * <code>RuleBase</code>.
     *
     * @return A newly initialized <code>WorkingMemory</code>.
     */
    public final WorkingMemory create(){

	return new WorkingMemory( this, this.transition );
    }

    /**
     * List of rule set names.  Returns null if empty.
     */
    public final String[] getNames(){

	int many = this.rulesets.size();

	if ( 0 < many){

	    String[] ret = new String[many];

	    this.rulesets.toArray(ret);

	    return ret;
	}
	else
	    return null;
    }


    /**
     * Assert a new fact object into this <code>RuleBase</code>
     * and the specified <code>WorkingMemory</code>.
     *
     * @param decl Fact identity.
     *
     * @param object Fact value.
     *
     * @param workingMemory The working memory session.
     *
     * @throws AssertionException if an error occurs during assertion.
     */
    protected final void assert( Declaration decl, Object object, WorkingMemory mem) throws AssertionException {

        this.rootnode.assert( decl, object, mem);
    }

    /**
     * Retract a fact object from this <code>RuleBase</code>
     * and the specified <code>WorkingMemory</code>.
     *
     * @param decl Fact identity.
     *
     * @param mem The working memory session.
     *
     * @throws RetractionException if an error occurs during retraction.
     */
    protected final void retract(Declaration decl, WorkingMemory mem) throws RetractionException {

        this.rootnode.retract( decl, mem);
    }

    /**
     * Modify a fact object in this <code>RuleBase</code> and the
     * specified <code>WorkingMemory</code>.
     *
     * <p> With the exception of time-based nodes, modification of a
     * fact object is semantically identical to retracting and
     * asserting it.
     *
     * @param decl Fact identity.
     *
     * @param object New fact value.
     *
     * @param mem The working memory session.
     *
     * @throws ModificationException if an error occurs during
     * assertion.
     */
    protected final void modify(Declaration decl, Object object, WorkingMemory mem) throws ModificationException {

        this.rootnode.modify( decl, object, mem);
    }
}
