/*
 * $Id$
 *
 * Copyright 2002 John Pritchard, All Rights Reserved.
 *
 * Licensed under the "Common Public License 1.0".
 *  http://www.yajre.org/license.txt
 */
/*
 * LICENSE.txt,v 1.1 2001/12/12 19:33:28 werken 
 *
 * Copyright 2001 (C) The Werken Company. All Rights Reserved.
 * 
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 * 
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 * 
 * 3. The name "Drools" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of The Werken Company.  For written permission,
 *    please contact bob@werken.com.
 * 
 * 4. Products derived from this Software may not be called "Drools"
 *    nor may "Drools" appear in their names without prior written
 *    permission of The Werken Company. Drools is a registered
 *    trademark of The Werken Company.
 * 
 * 5. Due credit should be given to the Drools Project
 *    (http://drools.org/).
 * 
 * THIS SOFTWARE IS PROVIDED BY THE WERKEN COMPANY AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE WERKEN COMPANY OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */

package org.yajre;

import org.yajre.reteoo.JoinMemory;
import org.yajre.reteoo.JoinNode;
import org.yajre.reteoo.Agenda;

import org.yajre.spi.Declaration;
import org.yajre.spi.ObjectType;
import org.yajre.spi.ActionInvokationException;

import java.util.Map;
import java.util.HashMap;
import java.util.Hashtable;

/**
 * A knowledge session for a {@link RuleBase}.
 *
 * @author <a href="mailto:bob@werken.com">bob@werken.com</a>
 * @author John Pritchard 
 */
public class WorkingMemory {

    /** The <code>RuleBase</code> with which this memory is associated. */
    private RuleBase rulebase;

    /** The actual memory for the {@link JoinNode}s. */
    private final Map joinMemories = new HashMap();

    /** Rule-firing agenda. */
    private Agenda   agenda;

    /**
     * Whether this session is holding actions.  
     */
    private boolean transition = false;//(jdp)

    /**
     * Allow integration components ("semantic") to attach state to
     * this session object.
     */
    private final Hashtable properties = new Hashtable();//(jdp)

    /**
     * External debugger.
     */
    private ChainListener debugger = null;

    protected WorkingMemory(RuleBase rulebase, boolean transition){
	super();

	if ( null == rulebase)

	    throw new IllegalArgumentException("Null argument.");

	else {
	    this.rulebase = rulebase;

	    this.agenda = new Agenda(this);

	    this.transition = transition;
	}
    }

    /**
     * Assign a chain listener when one is not set.  Use the unset
     * method before changing chain listeners.
     * 
     * @param lis Chain listener is typically some kind of debugger.
     * 
     * @see ChainListener
     */
    public final void setChainListener( ChainListener lis){
	if ( null == lis)
	    throw new IllegalArgumentException("Null argument, use the unset method.");

	else if ( null != this.debugger)

	    throw new IllegalArgumentException("Chain listener already set, use the unset method before resetting.");
	else
	    this.debugger = lis;
    }

    /**
     * Remove the current chain listener, for example before changing
     * chain listeners.
     * 
     * @see ChainListener
     */
    public final void unsetChainListener(){ this.debugger = null;}

    /**
     * Peek at the current chain listener.
     * 
     * @see ChainListener
     */
    public final ChainListener getChainListener(){ return this.debugger;}

    /**
     * Whether a chain listener is attached to this session.
     * 
     * @see ChainListener
     */
    public final boolean haveChainListener(){ return (null != this.debugger);}

    /**
     * Deallocate resources.
     */
    public void destroy(){

	if ( null != this.agenda){
	 
	    this.agenda.destroy();

	    this.agenda = null;
	}

	this.rulebase = null;

	this.joinMemories.clear();

	this.debugger = null;
    }

    public String toString(){
        return "WorkingMemory[" + this.joinMemories + "]";
    }

    /**
     * Whether this session is holding actions for explicit firing.
     * Default false.  When true, the "fire actions" method must be
     * called after assertions have been made, and when actions should
     * be fired.
     */
    public final boolean isTransition(){ return this.transition;}

    /**
     * Fire all actions collected into this agenda by fact updates.
     * This is used with the transition architecture -- otherwise,
     * actions are fired with each assertion.
     */
    public final void fireActions() throws ActionInvokationException {

	this.agenda.fireAll();
    }

    /**
     * Actions remain to be fired.  When "fire actions" throws an
     * exception, it should be reentered when actions remain to be
     * fired.
     */
    public final boolean haveActions(){

	return (!(getAgenda().isEmpty()));
    }

    /**
     * Allow integration components to attach state to this session
     * object.
     */
    public final Object getProperty ( Object key){

	Object val = this.properties.get(key);

	if ( null != val)

	    return val;
	else 
	    return this.rulebase.getAttribute(key);
    }
    /**
     * Allow integration components to attach state to this session
     * object.
     */
    public final void setProperty( Object key, Object value){

	this.properties.put(key,value);
    }
    /**
     * Allow integration components to attach state to this session
     * object.
     */
    public final void removeProperty( Object key){

	this.properties.remove(key);
    }

    /**
     * Lookup an attribute of this rule base.  (Used by integration
     * components for application purposes.)
     */
    public final Object getRuleBaseAttribute( Object key){

	return this.rulebase.getAttribute(key);
    }

    /**
     * Define an attribute of this rule base.  (Used by integration
     * components for application purposes.)
     */
    public final void setRuleBaseAttribute( Object key, Object value){

	this.rulebase.setAttribute(key,value);
    }

    /**
     * Delete an attribute of this rule base.  (Used by integration
     * components for application purposes.)
     */
    public final void removeRuleBaseAttribute( Object key){

	this.rulebase.removeAttribute(key);
    }

    /** 
     * Retrieve the rule-firing <code>Agenda</code> for this
     * <code>WorkingMemory</code>.
     *
     * @return The <code>Agenda</code>.
     */
    public final Agenda getAgenda(){

        return this.agenda;
    }

    /**
     * Return the names of rules ready to fire (in the agenda).
     */
    public final String[] getActionsRuleNames(){

	return getAgenda().ruleNames();
    }

    /** 
     * Retrieve the <code>RuleBase</code> of this working memory.
     *
     * @return The <code>RuleBase</code>.
     */
    public final RuleBase getRuleBase(){

        return this.rulebase;
    }

    /**
     * Given a parameter name, return its declaration if known.
     *
     * @returns Null if not found.
     */
    public final Declaration identity ( String id){

	return this.rulebase.identity(id);
    }



    /** 
     * Assert a new fact object into this working memory.
     *
     * @param fact Fact identifier.
     *
     * @param value Fact value.
     *
     * @throws AssertionException if an error occurs during assertion.
     */
    public final void assert( String fact, Object value) throws AssertionException {

	this.assert( this.rulebase.identity(fact), value);
    }

    /** 
     * Assert a new fact object into this working memory.
     *
     * @param decl Fact identity
     *
     * @param object The object to assert.
     *
     * @throws AssertionException if an error occurs during assertion.
     */
    public final void assert( Declaration decl, Object object) throws AssertionException {

	if ( null != decl && null != object){

	    getRuleBase().assert( decl, object, this );

	    if (!this.transition){

		Agenda agenda = getAgenda();

		while ( ! agenda.isEmpty() )

		    getAgenda().fireNextItem();
	    }
	}
    }

    /**
     * Retract a fact object from this working memory.
     *
     * @param fact Fact identity.
     *
     * @throws RetractionException if an error occurs during retraction.
     */
    public final void retract( String fact) throws RetractionException {

	this.retract( this.rulebase.identity(fact));
    }

    /**
     * Retract a fact object from this working memory.
     *
     * @param decl Fact identity.
     *
     * @throws RetractionException if an error occurs during retraction.
     */
    public final void retract( Declaration decl) throws RetractionException {

	if ( null != decl)
	    getRuleBase().retract( decl, this );
    }

    /**
     * Modify a fact object in this working memory.
     *
     * <p> With the exception of time-based nodes, modification of a
     * fact object is semantically equivelent to retracting and
     * re-asserting it.
     *
     * @param fact Fact identity.
     *
     * @param value Fact value.
     *
     * @throws ModificationException if an error occurs during modification.
     */
    public final void modify( String fact, Object value) throws ModificationException {

	this.modify( this.rulebase.identity(fact), value);
    }

    /**
     * Modify a fact object in this working memory.
     *
     * <p> With the exception of time-based nodes, modification of a
     * fact object is semantically equivelent to retracting and
     * re-asserting it.
     *
     * @param decl Fact identity
     *
     * @param object The object to modify.
     *
     * @throws ModificationException if an error occurs during modification.
     */
    public final void modify( Declaration decl, Object object) throws ModificationException {
	if ( null != decl && null != object)
	    getRuleBase().modify( decl, object, this );
    }



    /** Retrieve the {@link JoinMemory} for a particular {@link JoinNode}.
     *
     *  @param node The <code>JoinNode</code> key.
     *
     *  @return The node's memory.
     */
    public final JoinMemory getJoinMemory( JoinNode node){

        JoinMemory memory = (JoinMemory)this.joinMemories.get( node );

        if ( memory == null ){

            memory = new JoinMemory( node );

            this.joinMemories.put( node, memory );
        }

        return memory;
    }
}
