/*
 * $Id$
 *
 * Copyright 2002 John Pritchard, All Rights Reserved.
 *
 * Licensed under the "Common Public License 1.0".
 *  http://www.yajre.org/license.txt
 */
/*
 * LICENSE.txt,v 1.1 2001/12/12 19:33:28 werken 
 *
 * Copyright 2001 (C) The Werken Company. All Rights Reserved.
 * 
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 * 
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 * 
 * 3. The name "Drools" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of The Werken Company.  For written permission,
 *    please contact bob@werken.com.
 * 
 * 4. Products derived from this Software may not be called "Drools"
 *    nor may "Drools" appear in their names without prior written
 *    permission of The Werken Company. Drools is a registered
 *    trademark of The Werken Company.
 * 
 * 5. Due credit should be given to the Drools Project
 *    (http://drools.org/).
 * 
 * THIS SOFTWARE IS PROVIDED BY THE WERKEN COMPANY AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE WERKEN COMPANY OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */

package org.yajre.reteoo;

import org.yajre.spi.Rule;
import org.yajre.spi.Declaration;
import org.yajre.spi.ObjectType;
import org.yajre.spi.FilterCondition;
import org.yajre.spi.AssignmentCondition;
import org.yajre.spi.Condition;

import java.util.Set;
import java.util.HashSet;
import java.util.Iterator;

/**
 * "AddRule" is used by {@link org.yajre.RuleBase} to build the
 * Rete-OO network.  It is the implementation of C.L. Forgy's Rete
 * algorithm.
 *
 * @author <a href="mailto:bob@werken.com">bob@werken.com</a>
 * @author John Pritchard (john@syntelos.org)
 */
public abstract class Builder {

    /**
     * Add a {@link Rule} to the network.
     *
     * @param rule The rule to add.
     *
     * @throws ReteConstructionException if an error prevents complete
     *         construction of the network for the <code>Rule</code>.
     */
    public final static void AddRule(RootNode rootNode, Rule rule) throws ReteConstructionException {

        Set assignmentConds = new HashSet( rule.getAssignmentConditions());

        Set filterConds     = new HashSet( rule.getFilterConditions());

	/*
	 * Start the chain with the parameter nodes
	 */
        Set nodes = Link(rootNode, rule);

        boolean performedJoin     = false;

        boolean attachAssign      = false;

        boolean cycleAttachAssign = false;
        
	/*
	 * Fill- in the body of the chain
	 */
        do {
            performedJoin = false;

            attachAssign  = false;

            if (!filterConds.isEmpty())

                BooleanConditions( filterConds, nodes);

            attachAssign = ConsistentAssigns( assignmentConds, nodes);

            performedJoin = Join(nodes);
        }
        while ((!nodes.isEmpty())&&(performedJoin||attachAssign));

	/*
	 * Link the action to the bottom of the chain
	 */
	TupleSource lastNode = null;

	int many = nodes.size();

	if ( 1 < many){

	    Iterator nodesit = nodes.iterator();

	    while (nodesit.hasNext()){

		lastNode = (TupleSource)nodesit.next();

		if (lastNode instanceof ParameterNode)
		    
		    continue; // Skip extra declarations

		else          // (Filter, Assign, or Join)
		    break;
	    }
	}
	else if ( 1 == many)

	    lastNode = (TupleSource) nodes.iterator().next();


	if ( null != lastNode){

	    TerminalNode terminal = new TerminalNode(rule.getAction());

	    lastNode.setTupleSink(terminal);
	}
	else
	    throw new IllegalArgumentException("Incomplete rule `"+rule.getName()+"'.");
    }

    /** 
     * Create the {@link ParameterNode}s for the <code>Rule</code>,
     * and link into the network.
     *
     * @param rule The rule.
     *
     * @return A <code>Set</code> of <code>ParameterNodes</code>
     * created and linked into the network.
     */
    private final static Set Link(RootNode root, Rule rule){

        Set nodes = new HashSet();

        Iterator decls = rule.getParameterDeclarations().iterator();

        Declaration decl = null;

        while ( decls.hasNext() ){

            decl = (Declaration)decls.next();

            nodes.add(root.createParameterNode(rule,decl));
        }
        return nodes;
    }
    

    /** Create and attach {@link FilterCondition}s to the network.
     *
     *  <p>
     *  It may not be possible to satisfy all filder conditions
     *  on the first pass.  This method removes satisfied conditions
     *  from the <code>filterCond</code> parameter, and leaves
     *  unsatisfied ones in the <code>Set</code>.
     *  </p>
     *
     *  @param filterConds Set of <code>FilterConditions</code>
     *         to attempt attaching.
     *  @param nodes The current attachable leaf nodes
     *         of the network.
     */
    private final static void BooleanConditions(Set filterConds, Set nodes){

        Iterator        condIter    = filterConds.iterator();

        FilterCondition eachCond    = null;

        TupleSource     tupleSource = null;

        FilterNode filterNode = null;

        while ( condIter.hasNext() ){

            eachCond = (FilterCondition) condIter.next();

            tupleSource = BooleanConditionsSource( eachCond, nodes );

            if ( tupleSource == null )

                continue;

	    else {

		condIter.remove();

		filterNode = new FilterNode( tupleSource, eachCond );
		{
		    tupleSource.setTupleSink(filterNode);
		}
		nodes.remove( tupleSource );

		nodes.add( filterNode );
	    }
        }
    }

    /** Create and attach {@link JoinNode}s to the network.
     *
     *  <p>
     *  It may not be possible to join all <code>nodes</code>.
     *  </p>
     *
     *  <p>
     *  Any <code>attachabeNodes</code> member that particiates
     *  in a <i>join</i> is removed from the <code>nodes</code>
     *  collection, and replaced by the joining <code>JoinNode</code>.
     *  </p>
     *
     *  @param nodes The current attachable leaf nodes of
     *         the network.
     *
     *  @return <code>true</code> if at least one <code>JoinNode</code>
     *          was created, else <code>false</code>.
     */
    protected final static boolean Join(Set nodes){

        boolean performedJoin = false;

        Object[] leftNodes  = nodes.toArray();

        Object[] rightNodes = nodes.toArray();

        TupleSource left  = null;
        TupleSource right = null;

        JoinNode joinNode = null;

	outer:
        for ( int i = 0, j = 0; i < leftNodes.length ; ++i ){

            left = (TupleSource) leftNodes[i];

            if ( ! nodes.contains( left ) )

                continue outer;

	    else {

		inner:
		for ( j = i + 1; j < rightNodes.length ; ++j ){

		    right = (TupleSource) rightNodes[j];

		    if ( ! nodes.contains( right))

			continue inner;

		    else if ( CanBeJoined( left, right)){

			joinNode = new JoinNode( left, right );
			{
			    left.setTupleSink(joinNode);

			    right.setTupleSink(joinNode);
			}
			nodes.remove( left );

			nodes.remove( right );

			nodes.add( joinNode );

			performedJoin = true;

			continue outer;
		    }
		}
            }
        }

        return performedJoin;
    }

    /** Determine if two {@link TupleSource}s can be joined.
     *
     *  @return <code>true</code> if they can be joined (they share
     *          at least one common member declaration), else
     *          <code>false</code>.
     */
    private final static boolean CanBeJoined(TupleSource left, TupleSource right){

        Set      leftDecls     = left.getTupleDeclarations();

        Iterator rightDeclIter = right.getTupleDeclarations().iterator();

        while ( rightDeclIter.hasNext()){

            if ( leftDecls.contains( rightDeclIter.next()))

                return true;
        }

        return false;
    }

    /** Create and attach {@link AssignmentCondition}s to the network.
     *
     *  <p>
     *  It may not be possible to satisfy all <code>assignmentConds</code>,
     *  in which case, unsatisfied conditions will remain in the <code>Set</code>
     *  passed in as <code>assignmentConds</code>.
     *  </p>
     *
     *  @param assignmentConds Set of <code>AssignmentConditions</code> to
     *         attach to the network.
     *  @param nodes The current attachable leaf nodes of
     *         the network.
     */
    private final static boolean ConsistentAssigns(Set assignmentConds, Set nodes){
        boolean attached      = false;
        boolean cycleAttached = false;

        do {
            cycleAttached = false;

            Iterator            condIter  = assignmentConds.iterator();
            AssignmentCondition eachCond  = null;
            TupleSource         tupleSource = null;
            
            AssignmentNode assignNode = null;
            
            while ( condIter.hasNext()){

                eachCond = (AssignmentCondition) condIter.next();
                
                tupleSource = ConsistentAssignsSource( eachCond, nodes);
                
                if ( tupleSource == null )

                    continue;

                else {
		    condIter.remove();
                
		    assignNode = new AssignmentNode( tupleSource,
						     eachCond.getTargetDeclaration(),
						     eachCond.getFactExtractor() );
		    {
			tupleSource.setTupleSink(assignNode);
		    }
		    nodes.remove( tupleSource );
		    nodes.add( assignNode );
                
		    cycleAttached = true;
		}
	    }

            if (cycleAttached)

                attached = true;
        }
        while (cycleAttached);

        return attached;
    }

    /** Locate a <code>TupleSource</code> suitable for attaching
     *  the <code>FilterCondition</code>.
     *
     *  @param condition The <code>Condition</code> to attach.
     *  @param sources Candidate <code>TupleSources</code>.
     *
     *  @return Matching <code>TupleSource</code> if a suitable one
     *          can be found, else <code>null</code>.
     */
    private final static TupleSource BooleanConditionsSource(FilterCondition condition, Set sources){

        Iterator    sourceIter = sources.iterator();

        TupleSource eachSource = null;

        Set decls = null;

        while ( sourceIter.hasNext()){

            eachSource = (TupleSource) sourceIter.next();

            decls = eachSource.getTupleDeclarations();

            if ( Matches( condition, decls))

                return eachSource;
        }

        return null;
    }

    /** Locate a <code>TupleSource</code> suitable for attaching
     *  the <code>AssignmentCondition</code>.
     *
     *  @param condition The <code>Condition</code> to attach.
     *  @param sources Candidate <code>TupleSources</code>.
     *
     *  @return Matching <code>TupleSource</code> if a suitable one
     *          can be found, else <code>null</code>.
     */
    private final static TupleSource ConsistentAssignsSource(AssignmentCondition condition, Set sources){

        Declaration targetDecl = condition.getTargetDeclaration();

        Iterator    sourceIter = sources.iterator();

        TupleSource eachSource = null;

        Set decls = null;

        while ( sourceIter.hasNext() ){

            eachSource = (TupleSource) sourceIter.next();

            decls = eachSource.getTupleDeclarations();

            if ( decls.contains( targetDecl ) )

                continue;

            else if ( Matches( condition, decls))

                return eachSource;
        }

        return null;
    }

    /** Determine if a set of <code>Declarations</code> match those
     *  required by a <code>Condition</code>.
     *
     *  @param condition The <code>Condition</code>.
     *  @param declarations The set of <code>Declarations</code> to compare against.
     *
     *  @return <code>true</code> if the set of <code>Declarations</code> is a
     *          super-set of the <code>Declarations</code> required by the
     *          <code>Condition</code>.
     */
    private final static boolean Matches(Condition condition, Set declarations){

        Declaration[] requiredDecls = condition.getRequiredTupleMembers();

        for ( int i = 0 ; i < requiredDecls.length ; ++i ){

            if ( ! declarations.contains( requiredDecls[i]))

                return false;
        }

        return true;
    }
                                                  
}
