/*
 * $Id$
 *
 * Copyright 2002 John Pritchard, All Rights Reserved.
 *
 * Licensed under the "Common Public License 1.0".
 *  http://www.yajre.org/license.txt
 */
/*
 * LICENSE.txt,v 1.1 2001/12/12 19:33:28 werken 
 *
 * Copyright 2001 (C) The Werken Company. All Rights Reserved.
 * 
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 * 
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 * 
 * 3. The name "Drools" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of The Werken Company.  For written permission,
 *    please contact bob@werken.com.
 * 
 * 4. Products derived from this Software may not be called "Drools"
 *    nor may "Drools" appear in their names without prior written
 *    permission of The Werken Company. Drools is a registered
 *    trademark of The Werken Company.
 * 
 * 5. Due credit should be given to the Drools Project
 *    (http://drools.org/).
 * 
 * THIS SOFTWARE IS PROVIDED BY THE WERKEN COMPANY AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE WERKEN COMPANY OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */

package org.yajre.reteoo;

import org.yajre.spi.Declaration;

import java.util.List;
import java.util.LinkedList;
import java.util.ArrayList;
import java.util.Set;
import java.util.Iterator;
import java.util.Collections;

/**
 * Memory for left and right inputs of a {@link JoinNode}.
 *
 * @see JoinNode
 *
 * @see org.yajre.spi.Tuple
 *
 * @author <a href="mailto:bob@werken.com">bob@werken.com</a>
 * @author John Pritchard (john@syntelos.org)
 */
public final class JoinMemory {

    /**
     * <code>List</code> of <code>Tuples</code> received on the left
     * input of the <code>JoinNode</code>.
     */
    private final List leftTuples = new LinkedList();

    /**
     * <code>List</code> of <code>Tuples</code> received on the right
     * input of the <code>JoinNode</code>.
     */
    private final List rightTuples = new LinkedList();

    /**
     * The <code>Set</code> of common <code>Declarations</code> to
     * join against left and right memories.
     */
    private Set joinDeclarations;

    /** Construct.
     *
     *  @param node The <code>JoinNode</code> this memory
     *         is for.
     */
    public JoinMemory(JoinNode node){
	super();

        this.joinDeclarations = node.getCommonDeclarations();
    }

    /** Produce output suitable for debugging.
     */ 
    public String toString(){

        return "JoinMemory[left="+this.leftTuples+", right="+this.rightTuples+"]";
    }

    protected void retract( Declaration decl){

        ReteTuple each = null;

        Iterator  tuples = leftTuples.iterator();

        while ( tuples.hasNext()){

            each = (ReteTuple)tuples.next();

            if ( each.contains(decl))

                tuples.remove();
        }

        tuples = rightTuples.iterator();
        
        while ( tuples.hasNext()){

            each = (ReteTuple)tuples.next();

            if ( each.contains(decl))

                tuples.remove();
        }
    }

    /**
     * Add a {@link ReteTuple} received from the
     * <code>JoinNode's</code> left input to the left side of this
     * memory, and attempt to join to existing <code>Tuples</code> in
     * the right side.
     *
     * @param tuple The <code>Tuple</code> to add to the left side
     * memory.
     *
     * @return A <code>List</code> of <code>Tuples</code> successfully
     * created by joining the incoming <code>tuple</code> against
     * existing <code>Tuples</code> on the right side memory.
     */
    protected List addLeftTuple(ReteTuple tuple){

        this.leftTuples.add(tuple);

        return attemptJoin(tuple, getRightTupleIterator());
    }

    /** 
     * Retrieve the <code>List</code> of <code>Tuples</code> held in
     * the left side memory.
     *
     * @return The <code>List</code> of <code>Tuples</code> help in
     * the left side memory.
     */
    protected List getLeftTuples(){

        return this.leftTuples;
    }

    /**
     * Retrieve an <code>Iterator</code> over the <code>Tuples</code>
     * held in the left side memory.
     *
     * @return An <code>Iterator</code> over the <code>Tuples</code>
     * help in the left side memory.
     */
    protected Iterator getLeftTupleIterator(){

        return this.leftTuples.iterator();
    }

    /**
     * Add a {@link ReteTuple} received from the
     * <code>JoinNode's</code> right input to the right side of this
     * memory, and attempt to join to existing <code>Tuples</code> in
     * the left side.
     *
     * @param tuple The <code>Tuple</code> to add to the right side
     * memory.
     *
     * @return A <code>List</code> of <code>Tuples</code> successfully
     * created by joining the incoming <code>tuple</code> against
     * existing <code>Tuples</code> on the left side memory.
     */
    protected List addRightTuple(ReteTuple tuple){

        this.rightTuples.add( tuple);

        return attemptJoin( tuple, getLeftTupleIterator());
    }

    /**
     * Retrieve the <code>List</code> of <code>Tuples</code> held in
     * the right side memory.
     *
     * @return The <code>List</code> of <code>Tuples</code> help in
     * the right side memory.
     */
    protected List getRightTuples(){

        return this.rightTuples;
    }

    /**
     * Retrieve an <code>Iterator</code> over the <code>Tuples</code>
     * held in the right side memory.
     *
     * @return An <code>Iterator</code> over the <code>Tuples</code>
     * help in the right side memory.
     */
    protected Iterator getRightTupleIterator(){

        return this.rightTuples.iterator();
    }

    /**
     * Retrieve an <code>Iterator</code> over the common
     * <code>Declarations</code> used to join <code>Tuples</code> from
     * the left and right side memories.
     *
     * @return An <code>Iterator</code> of common join
     * <code>Declarations</code>.
     */
    protected Iterator getJoinDeclarationIterator(){

        return this.joinDeclarations.iterator();
    }

    /** 
     * Attempt to join the <code>tuple</code> against the tuples
     * available through the <code>tupleIterator</code>.
     *
     * @param tuple The <code>Tuple</code> to attempt joining.
     *
     * @param tupleIterator The <code>Iterator</code> over
     * <code>Tuples</code> to attempt joining to the
     * <code>tuple</code> parameter.
     *
     * @return A possibly empty <code>List</code> of joined
     * <code>Tuples</code>.
     */
    protected List attemptJoin(ReteTuple tuple, Iterator tupleIter){

        List joinedTuples = Collections.EMPTY_LIST;

        ReteTuple eachTuple   = null;
        ReteTuple joinedTuple = null;

        while ( tupleIter.hasNext()){

            eachTuple = (ReteTuple) tupleIter.next();
            
            joinedTuple = join( tuple, eachTuple);

            if ( joinedTuple != null ){

                if ( joinedTuples == Collections.EMPTY_LIST )

                    joinedTuples = new ArrayList();

                joinedTuples.add( joinedTuple );
            }
        }

        return joinedTuples;
    }

    /**
     * Attempt to join two <code>Tuples</code> (condense empty
     * fact slots).
     *
     * @param left The left-side <code>Tuple</code>.
     *
     * @param right The right-side <code>Tuple</code>.
     *
     * @return A newly joined <code>Tuple</code> if a join is
     * possible, else <code>null</code>.
     */
    private ReteTuple join(ReteTuple left, ReteTuple right){
        
        Iterator    declIter = getJoinDeclarationIterator();

        Declaration eachDecl = null;

        Object leftFact  = null;

        Object rightFact = null;

        while ( declIter.hasNext()){

            eachDecl = (Declaration)declIter.next();

            leftFact  = left.get( eachDecl );

            rightFact = right.get( eachDecl );

            if ( leftFact == null && rightFact == null)

                continue;

            else if ( leftFact == null || rightFact == null )

                return null;

            else if ( leftFact.equals( rightFact))

                continue;
            else
                return null;
        }

        ReteTuple joinedTuple = new JoinTuple( left, right );

        return joinedTuple;

    }
}
