/*
 * $Id$
 *
 * Copyright 2002 John Pritchard, All Rights Reserved.
 *
 * Licensed under the "Common Public License 1.0".
 *  http://www.yajre.org/license.txt
 */
/*
 * LICENSE.txt,v 1.1 2001/12/12 19:33:28 werken 
 *
 * Copyright 2001 (C) The Werken Company. All Rights Reserved.
 * 
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 * 
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 * 
 * 3. The name "Drools" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of The Werken Company.  For written permission,
 *    please contact bob@werken.com.
 * 
 * 4. Products derived from this Software may not be called "Drools"
 *    nor may "Drools" appear in their names without prior written
 *    permission of The Werken Company. Drools is a registered
 *    trademark of The Werken Company.
 * 
 * 5. Due credit should be given to the Drools Project
 *    (http://drools.org/).
 * 
 * THIS SOFTWARE IS PROVIDED BY THE WERKEN COMPANY AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE WERKEN COMPANY OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */

package org.yajre.reteoo;

import org.yajre.WorkingMemory;
import org.yajre.AssertionException;
import org.yajre.RetractionException;

import org.yajre.spi.Declaration;

import java.util.List;
import java.util.Set;
import java.util.HashSet;
import java.util.Iterator;

/**
 * A two-input Rete-OO <i>join node</i>.
 *
 * @see TupleSource
 * @see TupleSink
 *
 * @author <a href="mailto:bob@werken.com">bob@werken.com</a>
 * @author John Pritchard (john@syntelos.org)
 */
public final class JoinNode extends TupleSource implements TupleSink {
    /**
     * The left input <code>TupleSource</code>.
     */
    private TupleSource leftInput;

    /**
     * The right input <code>TupleSource</code>.
     */
    private TupleSource rightInput;

    /**
     * A <code>Set</code> of <code>Declarations</code> common to both
     * left and right input sources.
     */
    private Set commonDeclarations;

    /**
     * @param leftInput The left input <code>TupleSource</code>.
     *
     * @param rightInput The right input <code>TupleSource</code>.
     */
    public JoinNode(TupleSource left, TupleSource right){

        this.leftInput  = left;

        this.rightInput = right;

        this.commonDeclarations = _dintersect(left,right);
    }

    /**
     * Intersection of declarations of left and right sources.
     */
    private final static Set _dintersect(TupleSource left, TupleSource right){

	Set decls = new HashSet(left.getTupleDeclarations());

	decls.retainAll(right.getTupleDeclarations());

	return decls;
    }

    /**
     * Used by {@link JoinMemory}.  Return the intersection of left
     * and right declarations.
     */
    public Set getCommonDeclarations(){

        return this.commonDeclarations;
    }

    /**
     * Return the union of left and right declarations.
     */
    public Set getTupleDeclarations(){

        Set decls = new HashSet(getLeftInput().getTupleDeclarations());

        decls.addAll( getRightInput().getTupleDeclarations() );

        return decls;
    }

    /** Retrieve the left input <code>TupleSource</code>.
     *
     *  @return The left input <code>TupleSource</code>.
     */
    protected TupleSource getLeftInput(){

        return this.leftInput;
    }

    /** Retrieve the right input <code>TupleSource</code>.
     *
     *  @return The right input <code>TupleSource</code>.
     */
    protected TupleSource getRightInput(){

        return this.rightInput;
    }

    public void sinkAssertTuple(TupleSource src, ReteTuple tuple, WorkingMemory mem) throws AssertionException {

        JoinMemory memory = mem.getJoinMemory(this);

        List joinedTuples;

        if ( src == this.leftInput )

            joinedTuples = memory.addLeftTuple( tuple );
        else
            joinedTuples = memory.addRightTuple( tuple );

        if ( !joinedTuples.isEmpty() ){

	    Iterator tuples = joinedTuples.iterator();

	    ReteTuple each;

	    do {

		each = (ReteTuple)tuples.next();

		propagateAssertion( each, mem);
	    }
	    while ( tuples.hasNext());

	    return;
	}
    }

    public void sinkRetractObject(TupleSource src, Declaration decl, WorkingMemory mem) throws RetractionException {

        JoinMemory memory = mem.getJoinMemory(this);

        memory.retract(decl);
        
        propagateRetraction( decl, mem);
    }

    public String toString(){
	return "JoinNode[left="+this.leftInput+", right="+this.rightInput+"]";
    }
}
