/*
 * $Id$
 *
 * Copyright 2002 John Pritchard, All Rights Reserved.
 *
 * Licensed under the "Common Public License 1.0".
 *  http://www.yajre.org/license.txt
 */
/*
 * LICENSE.txt,v 1.1 2001/12/12 19:33:28 werken 
 *
 * Copyright 2001 (C) The Werken Company. All Rights Reserved.
 * 
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 * 
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 * 
 * 3. The name "Drools" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of The Werken Company.  For written permission,
 *    please contact bob@werken.com.
 * 
 * 4. Products derived from this Software may not be called "Drools"
 *    nor may "Drools" appear in their names without prior written
 *    permission of The Werken Company. Drools is a registered
 *    trademark of The Werken Company.
 * 
 * 5. Due credit should be given to the Drools Project
 *    (http://drools.org/).
 * 
 * THIS SOFTWARE IS PROVIDED BY THE WERKEN COMPANY AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE WERKEN COMPANY OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */

package org.yajre.reteoo;

import org.yajre.WorkingMemory;
import org.yajre.AssertionException;
import org.yajre.RetractionException;
import org.yajre.ModificationException;

import org.yajre.spi.ObjectType;
import org.yajre.spi.Declaration;

import java.util.* ;


/**
 * Filters <code>Objects</code> coming from the {@link RootNode} using
 * a {@link ObjectType} semantic module.
 *
 * <p> It receives <code>Objects</code> from the
 * <code>RootNode</code>, uses a <code>ObjectType</code> instance to
 * determine membership, and propagates matching <code>Objects</code>
 * further to all matching {@link ParameterNode}s.
 *
 * @see ObjectType
 * @see ParameterNode
 *
 * @author <a href="mailto:bob@werken.com">bob@werken.com</a>
 * @author John Pritchard (john@syntelos.org)
 */
public final class ObjectTypeNode {

    /**
     * The <code>ObjectType</code> semantic module. 
     */
    private ObjectType objectType;

    /** 
     * <code>ParameterNode</code> heads of their respective rule
     * chains.
     */
    private final List parameters = new ArrayList();

    /** 
     * <code>Declaration</code>s indexed by declaration identifier.
     */
    private final Map declarations = new HashMap();

    /** 
     * Construct given a semantic <code>ObjectType</code>.
     *
     * @param objectType The semantic object-type differentiator.
     */
    ObjectTypeNode(ObjectType objectType){
	super();

	if ( null == objectType)
	    throw new IllegalArgumentException("Null argument.");
	else
	    this.objectType = objectType;
    }

    public final Declaration identity( String parmid){

	return (Declaration)this.declarations.get(parmid);
    }

    /**
     * Retrieve the semantic <code>ObjectType</code> differentiator.
     *
     *  @return The semantic <code>ObjectType</code> differentiator.
     */
    public final ObjectType getObjectType(){

        return this.objectType;
    }

    /** Add a <code>ParameterNode</code> child to this node.
     *
     *  @param node The <code>ParameterNode</code> child to add.
     */
    protected final void addParameterNode(ParameterNode node){

	Declaration decl = node.getDeclaration();

        this.declarations.put(decl.getIdentifier(), decl);

        this.parameters.add(node);
    }

    /**
     * Assert a fact.
     *
     * @param decl Fact identity.
     *
     * @param object Fact value.
     *
     * @throws DroolsException if an error occurs during assertion.
     */
    public final void assert( Declaration decl, Object object, WorkingMemory mem) throws AssertionException {

	Iterator nodes = this.parameters.iterator();

	ParameterNode pn;

	while (nodes.hasNext()){

	    pn = (ParameterNode)nodes.next();

	    if (pn.getDeclaration().equals(decl))

		pn.assert( object, mem);
        }
    }

    /** Retract a fact object from this <code>RuleBase</code>
     *  and the specified <code>WorkingMemory</code>.
     *
     *  @param object The object to retract.
     *
     *  @param workingMemory The working memory session.
     *
     *  @throws DroolsException if an error occurs during assertion.
     */
    public final void retract( Declaration decl, WorkingMemory mem) throws RetractionException {

	Iterator nodes = this.parameters.iterator();

	ParameterNode pn = null;

	while (nodes.hasNext()){

	    pn = (ParameterNode)nodes.next();

	    if (pn.getDeclaration().equals(decl))

		pn.retract(mem);
	}
    }

    /** Modify a fact object in this <code>RuleBase</code>
     *  and the specified <code>WorkingMemory</code>.
     *
     *  With the exception of time-based nodes, modification of
     *  a fact object is semantically equivelent to retracting and
     *  re-asserting it.
     *
     *  @param object The object to modify.
     *  @param workingMemory The working memory session.
     *
     *  @throws DroolsException if an error occurs during assertion.
     */
    public final void modify( Declaration decl, Object object, WorkingMemory mem) throws ModificationException {

	Iterator nodes = this.parameters.iterator();

	ParameterNode pn = null;

	while (nodes.hasNext()){

	    pn = (ParameterNode)nodes.next();

	    if (pn.getDeclaration().equals(decl))

		pn.modify( object, mem);
        }
    }
}
