/*
 * $Id$
 *
 * Copyright 2002 John Pritchard, All Rights Reserved.
 *
 * Licensed under the "Common Public License 1.0".
 *  http://www.yajre.org/license.txt
 */
/*
 * LICENSE.txt,v 1.1 2001/12/12 19:33:28 werken 
 *
 * Copyright 2001 (C) The Werken Company. All Rights Reserved.
 * 
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 * 
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 * 
 * 3. The name "Drools" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of The Werken Company.  For written permission,
 *    please contact bob@werken.com.
 * 
 * 4. Products derived from this Software may not be called "Drools"
 *    nor may "Drools" appear in their names without prior written
 *    permission of The Werken Company. Drools is a registered
 *    trademark of The Werken Company.
 * 
 * 5. Due credit should be given to the Drools Project
 *    (http://drools.org/).
 * 
 * THIS SOFTWARE IS PROVIDED BY THE WERKEN COMPANY AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE WERKEN COMPANY OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */

package org.yajre.reteoo;

import org.yajre.WorkingMemory;
import org.yajre.AssertionException;
import org.yajre.RetractionException;
import org.yajre.ModificationException;

import org.yajre.spi.Declaration;

import java.util.* ;

/** Receives <code>Objects</code> from an {@link ObjectTypeNode},
 *  and creates a {@link ReteTuple}, passing the result to the following node.
 *
 *  <p>
 *  The <code>ParameterNode</code> is the first node that works in
 *  terms of <code>Tuples</code>.  An instance of <code>ParameterNode</code>
 *  exists for each <i>root fact object</i> parameter of each rule.
 *  </p>
 *
 *  @see ObjectTypeNode
 *  @see TupleSink
 *
 *  @author <a href="mailto:bob@werken.com">bob@werken.com</a>
 */
public final class ParameterNode extends TupleSource {

    private Declaration declaration;

    /** 
     *  @param inputNode The <code>ObjectTypeNode</code> input to this.
     *
     *  @param declaration The root fact object {@link Declaration}.
     */
    public ParameterNode( ObjectTypeNode inputNode, Declaration declaration){
	super();

        this.declaration = declaration;

        if ( inputNode != null )
            inputNode.addParameterNode( this );
    }

    /** 
     * The fact object <code>Declaration</code>.
     *
     *  @return The <code>Declaration</code>.
     */
    public final Declaration getDeclaration(){
        return this.declaration;
    }

    /**
     * @see TupleSource
     */
    public final Set getTupleDeclarations(){

        return Collections.singleton( this.declaration );
    }

    protected final void assert( Object obj, WorkingMemory mem) throws AssertionException {

        ReteTuple tuple = new ParameterTuple( this.declaration, obj);

        super.propagateAssertion( tuple, mem);
    }

    protected final void retract( WorkingMemory mem) throws RetractionException {

        super.propagateRetraction( this.declaration, mem);
    }

    protected final void modify( Object obj, WorkingMemory mem) throws ModificationException {
        try {
            super.propagateRetraction( this.declaration, mem);
            
            assert( obj, mem );
        }
        catch (RetractionException e){

            throw new ModificationException( e );
        }
        catch (AssertionException e){

            throw new ModificationException( e );
        }
    }

    public String toString(){

	return "ParameterNode["+this.declaration+"]";
    }
}
