/*
 * $Id$
 *
 * Copyright 2002 John Pritchard, All Rights Reserved.
 *
 * Licensed under the "Common Public License 1.0".
 *  http://www.yajre.org/license.txt
 */
/*
 * LICENSE.txt,v 1.1 2001/12/12 19:33:28 werken 
 *
 * Copyright 2001 (C) The Werken Company. All Rights Reserved.
 * 
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 * 
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 * 
 * 3. The name "Drools" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of The Werken Company.  For written permission,
 *    please contact bob@werken.com.
 * 
 * 4. Products derived from this Software may not be called "Drools"
 *    nor may "Drools" appear in their names without prior written
 *    permission of The Werken Company. Drools is a registered
 *    trademark of The Werken Company.
 * 
 * 5. Due credit should be given to the Drools Project
 *    (http://drools.org/).
 * 
 * THIS SOFTWARE IS PROVIDED BY THE WERKEN COMPANY AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE WERKEN COMPANY OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */

package org.yajre.reteoo;

import org.yajre.WorkingMemory;
import org.yajre.AssertionException;
import org.yajre.RetractionException;
import org.yajre.ModificationException;

import org.yajre.spi.Declaration;
import org.yajre.spi.ObjectType;
import org.yajre.spi.Rule;

import java.util.* ;

/**
 * The root node of the Rete-OO network is used by the {@link
 * org.yajre.RuleBase}.
 *
 * @author <a href="mailto:bob@werken.com">bob@werken.com</a>
 * @author John Pritchard (john@syntelos.org)
 */
public final class RootNode {

    /**
     * The set of <code>ObjectTypeNodes</code> indexed by
     * <code>ObjectType</code>.
     */
    private final Map objectTypeNodes = new HashMap();

    /**
     * Declarations indexed by identifier.
     */
    private final Map parameters = new HashMap();

    /** Construct.
     */
    public RootNode(){
	super();
    }

    /**
     * Given a parameter identifier, return its declaration, or null
     * if not found.
     */
    public final Declaration identity( String name){

	return (Declaration)this.parameters.get(name);
    }

    /**
     * Given an (user) type name or parameter identifier, return or
     * create its declaration.
     * 
     * <p> Rule set readers need to use this function (via {@link
     * org.yajre.RuleBase}), which resolves type issues.
     * 
     * @param type Fact identity type.
     * 
     * @param id Fact identity name.
     *
     * @returns Declaration
     */
    public final Declaration identity ( ObjectType type, String id){ 

	Declaration decl = identity(id);

	if ( null != decl){
	    /*
	     * Return the lowest common denominator (type)
	     */

	    ObjectType declt = decl.getObjectType();

	    if ( declt.equals(type))

		return decl;

	    else if (declt.subType(type)){

		return decl;
	    }
	    else if (type.subType(declt)){

		decl = new Declaration(type,id);

		this.parameters.put(id,decl);

		return decl;
	    }
	    else
		throw new IllegalArgumentException("Rules' parameter `"+id+"' defined with type `"+declt.getType()+"' being used with conflicting type `"+type.getType()+"'.");
	}
	else {

	    decl = new Declaration(type,id);

	    this.parameters.put(id,decl);

	    return decl;
	}
    }



    /**
     * Assert a new fact.
     *
     * @param decl Fact identity.
     *
     * @param object Fact value.
     *
     * @param mem The working memory session.
     *
     * @throws DroolsException if an error occurs during assertion.
     */
    public void assert(Declaration decl, Object object, WorkingMemory mem) throws AssertionException {
	
        ObjectTypeNode node = (ObjectTypeNode)this.objectTypeNodes.get(decl.getObjectType());

	if ( null != node)
	    node.assert( decl, object, mem);
    }

    /**
     * Retract a fact.
     *
     * @param decl Fact identity.
     *
     * @param mem The working memory session.
     *
     * @throws DroolsException if an error occurs during assertion.
     */
    public void retract( Declaration decl, WorkingMemory mem) throws RetractionException {

        ObjectTypeNode node = (ObjectTypeNode)this.objectTypeNodes.get(decl.getObjectType());

	if ( null != node)
	    node.retract( decl, mem);
    }

    /** 
     * Modify a fact.
     *
     * <p> With the exception of time-based nodes, modification of a
     * fact object is semantically identical to retracting and
     * asserting it.
     *
     * @param decl Fact identity.
     *
     * @param object Fact value.
     *
     * @param mem The working memory session.
     *
     * @throws DroolsException if an error occurs during assertion.
     */
    public void modify (Declaration decl, Object object, WorkingMemory mem) throws ModificationException {

        ObjectTypeNode node = (ObjectTypeNode)this.objectTypeNodes.get(decl.getObjectType());

	if ( null != node)
	    node.modify( decl, object, mem);
    }



    /**
     * Add an <code>ObjectTypeNode</code> child to this
     * network.
     *
     * @param node Node to add.
     */
    private final void add (ObjectTypeNode node){

	ObjectType ot = node.getObjectType();

        this.objectTypeNodes.put( ot, node );
    }

    /**
     * Retrieve an {@link ObjectTypeNode} keyed by {@link ObjectType}.
     *
     * @param ot The <code>ObjectType</code> key.
     *
     * @return The matching <code>ObjectTypeNode</code> if one has
     * already been created, else <code>null</code>.
     */
    protected final ObjectTypeNode getObjectTypeNode(ObjectType ot){

        return (ObjectTypeNode)this.objectTypeNodes.get(ot);
    }

    /**
     * Retrieve an {@link ObjectTypeNode} keyed by {@link ObjectType},
     * creating one, if necessary.
     *
     * @param ot The <code>ObjectType</code> key.
     *
     * @return The matching <code>ObjectTypeNode</code>.
     */
    protected final ObjectTypeNode getOrCreateObjectTypeNode(ObjectType ot){

        ObjectTypeNode node = getObjectTypeNode(ot);

        if ( node == null ){

            node = new ObjectTypeNode(ot);

            this.add(node);
        }

        return node;
    }

    /**
     * Used by Rete Builder to create a new rule (chain) head.
     */
    protected final ParameterNode createParameterNode( Rule rule, Declaration decl){

	ObjectType ot = decl.getObjectType();

	ObjectTypeNode otn = getOrCreateObjectTypeNode(ot);

	String parmid = decl.getIdentifier();

	if ( this.parameters.containsKey(parmid)){

	    Declaration chk = (Declaration)this.parameters.get( parmid);

	    if (!decl.equals(chk))
		throw new IllegalStateException("BBBUGGGG!  Parameter `"+parmid+"' has been declared with type `"+chk.getObjectType().getType()+"', used in rule `"+rule.getName()+"' with type `"+ot.getType()+"'.");
	}
	else 
	    this.parameters.put( parmid, decl);

	ParameterNode paramNode = new ParameterNode( otn, decl);

	otn.addParameterNode(paramNode);
	    
	return paramNode;
    }
}
