/*
 * $Id$
 *
 * Copyright 2002 John Pritchard, All Rights Reserved.
 *
 * Licensed under the "Common Public License 1.0".
 *  http://www.yajre.org/license.txt
 */
/*
 * LICENSE.txt,v 1.1 2001/12/12 19:33:28 werken 
 *
 * Copyright 2001 (C) The Werken Company. All Rights Reserved.
 * 
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 * 
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 * 
 * 3. The name "Drools" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of The Werken Company.  For written permission,
 *    please contact bob@werken.com.
 * 
 * 4. Products derived from this Software may not be called "Drools"
 *    nor may "Drools" appear in their names without prior written
 *    permission of The Werken Company. Drools is a registered
 *    trademark of The Werken Company.
 * 
 * 5. Due credit should be given to the Drools Project
 *    (http://drools.org/).
 * 
 * THIS SOFTWARE IS PROVIDED BY THE WERKEN COMPANY AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE WERKEN COMPANY OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */

package org.yajre.reteoo;

import org.yajre.WorkingMemory;
import org.yajre.AssertionException;
import org.yajre.RetractionException;

import org.yajre.spi.Declaration;

import java.util.Set;

/**
 * A source of {@link ReteTuple}s for a {@link TupleSink}.
 *
 * <p> Chain nodes propagate <code>Tuples</code> and extend this
 * class.
 *
 * @see TupleSink
 *
 * @author <a href="mailto:bob@werken.com">bob@werken.com</a>
 * @author John Pritchard (john@syntelos.org)
 */
public abstract class TupleSource {

    /**
     * The next link in the chain.
     */
    private TupleSink tupleSink;

    /** 
     *
     */
    protected TupleSource(){
	super();
    }

    /** Retrieve the <code>Set</code> of {@link Declaration}s
     *  in the propagated <code>Tuples</code>.
     *
     *  @return The <code>Set</code> of <code>Declarations</code>
     *          in progated <code>Tuples</code>.
     */
    protected abstract Set getTupleDeclarations();

    /**
     * Set the <code>TupleSink</code> that receives
     * <code>Tuples</code> propagated from this
     * <code>TupleSource</code>.
     *
     * @param tupleSink The <code>TupleSink</code> to receive
     * propagated <code>Tuples</code>.
     */
    protected final void setTupleSink(TupleSink tupleSink){

	if ( null != this.tupleSink)
	    throw new IllegalStateException("BBBUGGGG!  Overwriting tuple sink.");
	else
	    this.tupleSink = tupleSink;
    }

    /**
     * Retrieve the <code>TupleSink</code> that receives propagated
     * <code>Tuples</code>.
     *
     * @return The <code>TupleSink</code> that receives propagated
     * <code>Tuples</code>.
     */
    protected final TupleSink getTupleSink(){

        return this.tupleSink;
    }

    /** 
     * Propagate the assertion of a <code>Tuple</code> to the next
     * <code>TupleSink</code> node in the chain.
     *
     * @param tuple The asserted <code>Tuple</code> to propagate.
     *
     * @param mem The working memory session.
     */
    protected final void propagateAssertion(ReteTuple tuple, WorkingMemory mem) throws AssertionException {

        TupleSink sink = getTupleSink();

	if (mem.haveChainListener())
	    mem.getChainListener().chainStepAssertion(this,tuple,sink);

        if ( sink != null )
            sink.sinkAssertTuple( this, tuple, mem);
    }

    /**
     * Propagate the retraction of a <code>Tuple</code> to the next
     * <code>TupleSink</code> node in the chain.
     *
     * @param decl The retracted fact identity.
     *
     * @param mem The working memory session.
     */
    protected final void propagateRetraction( Declaration decl, WorkingMemory mem) throws RetractionException {

        TupleSink sink = getTupleSink();

	if (mem.haveChainListener())
	    mem.getChainListener().chainStepRetraction(this,decl,sink);

        if ( sink != null )
            sink.sinkRetractObject( this, decl, mem);
    }
}
