/*
 * $Id$
 *
 * Copyright 2002 John Pritchard, All Rights Reserved.
 *
 * Licensed under the "Common Public License 1.0".
 *  http://www.yajre.org/license.txt
 */

package org.yajre.semantic.java;

import bsh.Interpreter;
import bsh.EvalError;

import org.yajre.* ;

import org.yajre.reteoo.TupleSource;
import org.yajre.reteoo.ReteTuple;
import org.yajre.reteoo.TupleSink;

import org.yajre.spi.Declaration;

import java.io.* ;

import java.util.* ;

/**
 * Console rules debugger used by other console programs,
 * <code>`test'</code> and <code>`bshi'.</code>
 * 
 * @author John Pritchard
 */
public final class db implements ChainListener {

    private final static String bsh_function_dp = 
	"dp(){"+
	" out.println(\"p> \"+src);"+
	" switch(mode){"+
	" case 1:"+
	"  out.println(\"p> \"+tuple);"+
	"  break;"+
	" case 2:"+
	"  out.println(\"p> \"+decl);"+
	"  break;"+
	" }"+
	" out.println(\"p> \"+sink);"+
	"}";

    private final static int MODE_NULL = 0;
    private final static int MODE_ASSERT = 1;
    private final static int MODE_RETRACT = 2;

    private final static String prompt_null = "debug> ";
    private final static String prompt_assert = "assert-debug> ";
    private final static String prompt_retrac = "retrac-debug> ";

    public final static void help( int mode, PrintStream out){
	out.println();
	out.println(" [Command]            -- [Description]");
	out.println();
	out.println(" g <name>             -- Get name, print value.");
	out.println();
	out.println(" s <name> <value>     -- Set name, print value.");
	out.println();
	out.println(" u <name>             -- Unset name.");
	out.println();
	out.println(" (eval)               -- Evaluate `bsh' input.");
	out.println();
	out.println(" q                    -- Quit, allow rules to proceed.");
	out.println();
	out.println(" ?|h                  -- This `help' message.");
	out.println();
	out.println(" [Variable name]      -- [Description]");
	out.println();
	out.println(" mode                 -- Integer assert=1, retract=2.");
	out.println();
	out.println(" out                  -- Console output stream.");
	out.println();
	out.println(" err                  -- Console error stream.");
	out.println();
	switch(mode){
	case MODE_ASSERT:
	    out.println(" src                  -- Source.");
	    out.println();
	    out.println(" tuple                -- Tuple.");
	    out.println();
	    out.println(" sink                 -- Sink.");
	    out.println();
	    break;
	case MODE_RETRACT:
	    out.println(" src                  -- Source.");
	    out.println();
	    out.println(" decl                 -- Declaration.");
	    out.println();
	    out.println(" sink                 -- Sink.");
	    out.println();
	    break;
	}
	out.println(" [Function]           -- [Description]");
	out.println();
	out.println(" dp();                 -- Print src, (tuple|decl), sink.");
	out.println();
    }

    public Interpreter interp;

    public DataInputStream console_in;

    public PrintStream console_out;

    public PrintStream console_err;

    private int mode = MODE_NULL;

    public db( Interpreter bsi, DataInputStream in, PrintStream out, PrintStream err){
	super();

	if ( null == bsi)
	    throw new IllegalArgumentException("Null interpreter.");
	else
	    this.interp = bsi;

	if ( null == in)
	    throw new IllegalArgumentException("Null input stream.");
	else
	    this.console_in = in;

	if ( null == out)
	    throw new IllegalArgumentException("Null output stream.");
	else
	    this.console_out = out;

	if ( null == err)
	    throw new IllegalArgumentException("Null error stream.");
	else
	    this.console_err = err;

	try {
	    this.interp.eval(bsh_function_dp);
	}
	catch ( EvalError ever){

	    ever.printStackTrace();
	}
    }

    public void chainStepAssertion( TupleSource src, ReteTuple tuple, TupleSink sink){

	this.mode = MODE_ASSERT;
	try {
	    this.interp.set("src",src);
	    this.interp.set("tuple",tuple);
	    this.interp.set("sink",sink);
	    this.interp.set("mode",new Integer(this.mode));
	    this.interp.set("out",this.console_out);
	    this.interp.set("err",this.console_err);

	    this.console();
	}
	catch ( Exception exc){
	    exc.printStackTrace();
	}
	finally {
	    this.mode = MODE_NULL;
	    try {
		this.interp.unset("src");
		this.interp.unset("tuple");
		this.interp.unset("sink");
		this.interp.unset("mode");
		this.interp.unset("out");
		this.interp.unset("err");
	    }
	    catch ( EvalError ever){}
	}
    }

    public void chainStepRetraction( TupleSource src, Declaration decl, TupleSink sink){

	this.mode = MODE_RETRACT;
	try {
	    this.interp.set("src",src);
	    this.interp.set("decl",decl);
	    this.interp.set("sink",sink);
	    this.interp.set("mode",new Integer(this.mode));
	    this.interp.set("out",this.console_out);
	    this.interp.set("err",this.console_err);

	    this.console();
	}
	catch ( Exception exc){
	    exc.printStackTrace();
	}
	finally {
	    this.mode = MODE_NULL;
	    try {
		this.interp.unset("src");
		this.interp.unset("decl");
		this.interp.unset("sink");
		this.interp.unset("mode");
		this.interp.unset("out");
		this.interp.unset("err");
	    }
	    catch ( EvalError ever){}
	}
    }


    public final Object eval( String input, PrintStream out) throws EvalError {

	Object value = this.interp.eval(input);

	if ( null != value)
	    out.println(value);
	else
	    out.println();

	return value;
    }

    private final void console() throws IOException {

	this.console(console_in,console_out,console_err);
    }

    public final void console ( DataInputStream din, PrintStream out, PrintStream err) throws IOException {

	String line, name;

	Object value;

	int linelen;

	StringTokenizer strtok;

	switch(this.mode){
	case MODE_ASSERT:
	    err.print(prompt_assert);
	    break;
	case MODE_RETRACT:
	    err.print(prompt_retrac);
	    break;
	default:
	    err.print(prompt_null);
	    break;
	}

	while (null != (line = din.readLine())){

	    try {

		linelen = line.length();

		if ( 0 < linelen){

		    switch(line.charAt(0)){

			/*
			 * Get <name>
			 */
		    case 'g':
			if ( 1 < linelen && ' ' == line.charAt(1)){
			    strtok = new StringTokenizer(line.substring(1)," ");

			    if (strtok.hasMoreTokens()){

				name = strtok.nextToken();

				value = this.interp.get(name);

				out.println(value);

				break;
			    }
			    else
				throw new IllegalArgumentException("Command 'g'et requires argument `name'.");
			}
			else
			    this.eval(line,out);

			break;

			/*
			 * Set <name> <value>
			 */
		    case 's':
			if ( 1 < linelen && ' ' == line.charAt(1)){
			    strtok = new StringTokenizer(line.substring(1)," ");

			    if (strtok.hasMoreTokens()){

				name = strtok.nextToken();

				if (strtok.hasMoreTokens()){

				    value = strtok.nextToken();

				    this.interp.set(name,value);

				    out.println(value);

				    break;
				}
			    }

			    throw new IllegalArgumentException("Command 's'et requires arguments `name' and `value'.");
			}
			else
			    this.eval(line,out);

			break;

			/*
			 * Unset <name>
			 */
		    case 'u':
			if ( 1 < linelen && ' ' == line.charAt(1)){
			    strtok = new StringTokenizer(line.substring(1)," ");

			    if (strtok.hasMoreTokens()){

				name = strtok.nextToken();

				this.interp.unset(name);

				out.println();

				break;
			    }
			    else
				throw new IllegalArgumentException("Command 'u'nset requires argument `name'.");
			}
			else
			    this.eval(line,out);

			break;

		    case '?':
		    case 'h':

			if ( 1 == linelen)

			    help(this.mode,out);
			else
			    this.eval(line,out);

			break;

		    case 'q':

			if ( 1 == linelen)

			    return;
			else
			    this.eval(line,out);

			break;

		    default:

			this.eval(line,out);

			break;
		    }
		}
	    }
	    catch ( IllegalArgumentException ilarg){

		String msg = ilarg.getMessage();

		if ( null == msg)
		    help(this.mode,out);
		else
		    out.println("Error: "+msg);

	    }
	    catch ( EvalError ever){

		String msg = ever.getMessage();
		try {
		    int idx = msg.indexOf('>');

		    if ( 0 < idx)
			msg = msg.substring(idx+3).trim();

		} catch ( Exception x){}

		out.println("[error] "+msg);
	    }
	    catch ( Exception exc){

		out.print("[error] ");

		exc.printStackTrace(out);
	    }

	    switch(this.mode){
	    case MODE_ASSERT:
		err.print(prompt_assert);
		break;
	    case MODE_RETRACT:
		err.print(prompt_retrac);
		break;
	    default:
		err.print(prompt_null);
		break;
	    }

	}
    }

}
