/*
 * $Id$
 *
 * Copyright 2002 John Pritchard, All Rights Reserved.
 *
 * Licensed under the "Common Public License 1.0".
 *  http://www.yajre.org/license.txt
 */
/*
 * LICENSE.txt,v 1.1 2001/12/12 19:33:28 werken 
 *
 * Copyright 2001 (C) The Werken Company. All Rights Reserved.
 * 
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 * 
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 * 
 * 3. The name "Drools" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of The Werken Company.  For written permission,
 *    please contact bob@werken.com.
 * 
 * 4. Products derived from this Software may not be called "Drools"
 *    nor may "Drools" appear in their names without prior written
 *    permission of The Werken Company. Drools is a registered
 *    trademark of The Werken Company.
 * 
 * 5. Due credit should be given to the Drools Project
 *    (http://drools.org/).
 * 
 * THIS SOFTWARE IS PROVIDED BY THE WERKEN COMPANY AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE WERKEN COMPANY OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */

package org.yajre.semantic.java.io;

import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;

/**
 * 
 * @author Bob McWinter
 * @author John Pritchard
 */
public final class ImportManager {

    private List packages;

    private Map  classes;

    public ImportManager(){
	super();

        this.reset();
    }

    private final static String[] append( String[] ary, String el){
	if ( null == el)
	    return ary;
	else {
	    int idx = 0;

	    if (null == ary)
		ary = new String[1];
	    else {
		idx = ary.length;
		String[] copier = new String[idx+1];
		System.arraycopy(ary,0,copier,0,idx);
		ary = copier;
	    }
	    ary[idx] = el;
	    return ary;
	}
    }

    /**
     * Return a list of Java Programming Language import statements as
     * known by XML declared IMPORTs.  The output is a possibly empty
     * list of class imports, followed by a possibly empty list of
     * package imports.
     * 
     * <p> This value becomes the value of the
     * <code>"org.yajre.semantic.java.imports"</code> rule base
     * attribute.
     * 
     * @return Null if empty.
     */
    public String[] importStatements(){

	String[] ret = null;

	if ( 0 < classes.size()){

	    Iterator clss = classes.keySet().iterator();

	    String cla;

	    while (clss.hasNext()){

		cla = (String)clss.next();

		ret = append(ret,"import "+cla+" ;");
	    }
	} 

	if ( 0 < packages.size()){

	    Iterator pkgs = packages.iterator();

	    String pkg;

	    while (pkgs.hasNext()){

		pkg = (String)pkgs.next();

		ret = append(ret,"import "+pkg+"* ;");
	    }
	}

	return ret;
    }

    public void reset(){

        this.packages = new ArrayList();

        this.classes  = new HashMap();

        addPackageImport( "java.lang." );
    }

    public void addImport(String spec){
        if ( spec.endsWith( "*" ) )

            addPackageImport( spec.substring( 0, spec.length()-1));
        else
            addClassImport( spec );
    }

    protected void addPackageImport(String pkg){

        this.packages.add( pkg );
    }

    protected List getPackages(){

        return this.packages;
    }

    protected void addClassImport(String clazz){

        int lastDot = clazz.lastIndexOf( "." );

        String className = null;

        if ( lastDot >= 0 )

            className = clazz.substring( lastDot + 1 );
        else
            className = clazz;

        this.classes.put( className, clazz);
    }

    private final static Hashtable primitives = new Hashtable();
    static {
	primitives.put("boolean",Boolean.TYPE);
	primitives.put("byte",Byte.TYPE);
	primitives.put("short",Short.TYPE);
	primitives.put("char",Character.TYPE);
	primitives.put("int",Integer.TYPE);
	primitives.put("float",Float.TYPE);
	primitives.put("long",Long.TYPE);
	primitives.put("double",Double.TYPE);
	primitives.put("void",Void.TYPE);
    }

    public Class checkPrimitives(String name){

	return (Class)primitives.get(name);
    }

    public Class resolveClass(String name) throws ClassNotFoundException {

        Class clazz = null;

        clazz = checkPrimitives( name );

        if ( clazz != null )

            return clazz;


        String  fullName = null;

        if ( name.indexOf(".") >= 0 )

            fullName = name;
        else
            fullName = (String)this.classes.get(name);


        if ( fullName != null )

            clazz = Class.forName( fullName );

        else {

            Iterator pkgIter = getPackages().iterator();

            String   eachPkg = null;

            while ( pkgIter.hasNext() ){

                eachPkg = (String) pkgIter.next();

                fullName = eachPkg + name;

                try {

                    clazz = Class.forName( fullName );
                    
                    if ( clazz != null )

                        break;
                }
                catch (ClassNotFoundException e){}
            }
                
        }

        if ( clazz == null )

            throw new ClassNotFoundException(name);
	else 
	    return clazz;
    }
}

