/*
 * $Id$
 *
 * Copyright 2002 John Pritchard, All Rights Reserved.
 *
 * Licensed under the "Common Public License 1.0".
 *  http://www.yajre.org/license.txt
 */
/*
 * LICENSE.txt,v 1.1 2001/12/12 19:33:28 werken 
 *
 * Copyright 2001 (C) The Werken Company. All Rights Reserved.
 * 
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 * 
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 * 
 * 3. The name "Drools" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of The Werken Company.  For written permission,
 *    please contact bob@werken.com.
 * 
 * 4. Products derived from this Software may not be called "Drools"
 *    nor may "Drools" appear in their names without prior written
 *    permission of The Werken Company. Drools is a registered
 *    trademark of The Werken Company.
 * 
 * 5. Due credit should be given to the Drools Project
 *    (http://drools.org/).
 * 
 * THIS SOFTWARE IS PROVIDED BY THE WERKEN COMPANY AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE WERKEN COMPANY OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */

package org.yajre.semantic.java.io;

import org.yajre.RuleBase;

import org.yajre.spi.* ;

import org.dom4j.Document;
import org.dom4j.DocumentException;
import org.dom4j.io.SAXReader;

import java.io.File;
import java.net.URL;
import java.net.MalformedURLException;

import java.util.Map;
import java.util.HashMap;

/**
 * Reads Java XML ruleset files, producing a {@link RuleSet}.
 *
 * @author <a href="mailto:bob@werken.com">bob@werken.com</a>
 * @author John Pritchard
 */
public final class RuleSetReader {

    private RuleBase rulebase;

    private SAXReader reader = new SAXReader();

    private ImportManager importManager = new ImportManager();

    private RuleSet ruleSet;

    private Rule currentRule;

    /**
     * Construct.
     *
     * <p> A <code>RuleSetReader</code> may be re-used repeatedly, but
     * may <b>not</b> be used by two threads simultaneously.  It is
     * not synchronized nor re-entrant.
     */
    public RuleSetReader(RuleBase rulebase){

	if ( null == rulebase)

	    throw new IllegalArgumentException("BBBUGGGG! Null rule base arg.");
	else
	    this.rulebase = rulebase;

        this.reader.addHandler( "/ruleset",
                                new RuleSetHandler( this ) );

        this.reader.addHandler( "/ruleset/import",
                                new ImportHandler( this ) );

        this.reader.addHandler( "/ruleset/rule",
                                new RuleHandler( this ) );

        this.reader.addHandler( "/ruleset/rule/param",
                                new ParamHandler( this ) );

        this.reader.addHandler( "/ruleset/rule/decl",
                                new DeclHandler( this ) );

        this.reader.addHandler( "/ruleset/rule/when/cond",
                                new CondHandler( this ) );

        this.reader.addHandler( "/ruleset/rule/then",
                                new ThenHandler( this ) );
    }

    /**
     * Release references
     */
    public void destroy(){

	this.rulebase = null;
	this.reader = null;
	this.importManager = null;
	this.ruleSet = null;
	this.currentRule = null;
    }

    Declaration getDeclaration(String identifier){

        return this.rulebase.identity( identifier );
    }

    Declaration getDeclaration( ObjectType ot, String identifier){

        return this.rulebase.identity( ot, identifier );
    }

    Rule getCurrentRule(){

        return this.currentRule;
    }

    void setCurrentRule(Rule currentRule){

        this.currentRule = currentRule;
    }

    public ImportManager getImportManager(){

        return this.importManager;
    }

    void setRuleSet(RuleSet ruleSet){

        this.ruleSet = ruleSet;
    }

    RuleSet getRuleSet(){

        return this.ruleSet;
    }

    /** Read a Java ruleset XML file from a URL.
     *
     *  @param url The URL of the XML ruleset.
     *
     *  @return The resulting {@link RuleSet}.
     *
     *  @throws RuleSetReaderException if an error occurs
     *          while reading or parsing the file, or
     *          constructing the ruleset or rules.
     */
    public RuleSet read(URL url) throws RuleSetReaderException {

        getImportManager().reset();

        setCurrentRule( null );
        setRuleSet( null );

        try {

            Document doc = this.reader.read( url );
        }
        catch (DocumentException e){

	    e.printStackTrace();

            throw new RuleSetReaderException( e );
        }

        return getRuleSet();
    }

    /** Read a Java ruleset XML file from a local file.
     *
     *  @param file The File of the XML ruleset.
     *
     *  @return The resulting {@link RuleSet}.
     *
     *  @throws RuleSetReaderException if an error occurs
     *          while reading or parsing the file, or
     *          constructing the ruleset or rules.
     */
    public RuleSet read(File file) throws MalformedURLException, RuleSetReaderException {
        return read( file.toURL() );
    }
}
