/*
 * $Id$
 *
 * Copyright 2002 John Pritchard, All Rights Reserved.
 *
 * Licensed under the "Common Public License 1.0".
 *  http://www.yajre.org/license.txt
 */
/*
 * LICENSE.txt,v 1.1 2001/12/12 19:33:28 werken 
 *
 * Copyright 2001 (C) The Werken Company. All Rights Reserved.
 * 
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 * 
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 * 
 * 3. The name "Drools" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of The Werken Company.  For written permission,
 *    please contact bob@werken.com.
 * 
 * 4. Products derived from this Software may not be called "Drools"
 *    nor may "Drools" appear in their names without prior written
 *    permission of The Werken Company. Drools is a registered
 *    trademark of The Werken Company.
 * 
 * 5. Due credit should be given to the Drools Project
 *    (http://drools.org/).
 * 
 * THIS SOFTWARE IS PROVIDED BY THE WERKEN COMPANY AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE WERKEN COMPANY OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */

package org.yajre.spi;

import java.util.Set;
import java.util.HashSet;
import java.util.Collections;

/**
 * A set of {@link Condition}s and an {@link Action}.
 *
 * <p> The <code>Conditions</code> describe the circumstances that
 * represent a match for this rule.  The <code>Action</code> gets
 * fired when the <code>Conditions</code> match.
 *
 * @author <a href="mailto:bob@werken.com">bob@werken.com</a>
 * @author John Pritchard (john@syntelos.org)
 */
public class Rule {

    /**
     * Internal control for rule components' ordering: declarations
     * before conditions.
     */
    private boolean decl = true;

    private String name = null;

    private Set parameterDeclarations = Collections.EMPTY_SET;

    private Set localDeclarations = Collections.EMPTY_SET;

    private Set filterConditions = Collections.EMPTY_SET;

    private Set assignmentConditions = Collections.EMPTY_SET;

    private Action action = null;

    /** Construct.
     *
     *  @param name The name of this rule.
     */
    public Rule(String name){
	super();

        this.name  = name;
    }

    /** Produce output suitable for debugging.
     */
    public String toString(){

        return "Rule[name='" + this.name
            + "',paramDecls=" + this.parameterDeclarations
            + ",localDecls=" + this.localDeclarations
            + ",filterConds=" + this.filterConditions
            + ",assignConds=" + this.assignmentConditions
            + "]";
    }

    /** Determine if this rule is internally consistent and valid.
     *
     *  @return <code>true</code> if this rule is valid, else <code>false</code>.
     */
    public boolean isValid(){

        return ((!getParameterDeclarations().isEmpty())&&
		(!(getFilterConditions().isEmpty() &&
		   getAssignmentConditions().isEmpty())));
    }

    /** Check the validity of this rule, and throw exceptions if
     *  it failed validity tests.
     *
     *  <p>
     *  Possibly exceptions include:
     *  </p>
     * 
     *  <pre>
     *      NoParameterDeclarationException
     *      NoConditionException
     *  </pre>
     *
     *  <p>
     *  A <code>Rule</code> must include at least one
     *  parameter declaration and one condition.
     *  </p>
     *
     *  @throws InvalidRuleException if this rule is in any way invalid.
     */
    public void checkValidity() throws InvalidRuleException {

        if ( getParameterDeclarations().isEmpty() )

            throw new NoParameterDeclarationException( this );

        else if ( getFilterConditions().isEmpty() &&
		  getAssignmentConditions().isEmpty() )

            throw new NoConditionException( this );
    }

    /** Retrieve the name of this rule.
     *
     *  @return The name of this rule.
     */
    public String getName(){

        return this.name;
    }
    public void setName( String name){

	if ( null == this.name)

	    this.name = name;
	else
	    throw new IllegalStateException("Rule name already set.");
    }

    /** Add a <i>root fact object</i> parameter <code>Declaration</code>
     *  for this <code>Rule</code>.
     *
     *  <p>
     *  When constructing a rule, all parameter declarations must be
     *  specified before specifying any conditions.  Failure to do so
     *  may result in an invalid rule.  Attempting to add parameter
     *  declarations after a condition has been added will result in
     *  a <code>DeclarationAlreadyCompleteException</code>.
     *
     *  @param declaration The <i>root fact object</i> <code>Declaration</code> to add.
     *
     *  @throws DeclarationAlreadyCompleteException if any <code>Conditions</code> have
     *          already been added to this <code>Rule</code>.
     */
    public void addParameterDeclaration(Declaration declaration) throws DeclarationAlreadyCompleteException {

        if ( this.decl ){

            if ( this.parameterDeclarations == Collections.EMPTY_SET )

                this.parameterDeclarations = new HashSet();

            
            this.parameterDeclarations.add( declaration );
        }
        else
            throw new DeclarationAlreadyCompleteException(this);
    }

    private void addLocalDeclaration(Declaration declaration){

        this.decl = false;

        if ( this.localDeclarations == Collections.EMPTY_SET )

            this.localDeclarations = new HashSet();

        this.localDeclarations.add( declaration );
    }

    /** Retrieve the set of all <i>root fact object</i> parameter
     *  <code>Declarations</code>.
     *
     *  @return The <code>Set</code> of <code>Declarations</code> which
     *          specify the <i>root fact objects</i>.
     */
    public Set getParameterDeclarations(){

        return this.parameterDeclarations;
    }

    /** Retrieve the set of all implied local Declarations.
     *
     *  @return The <code>Set</code> of all implied <code>Declarations</code>
     *          which are implied by the conditions.
     */
    public Set getLocalDeclarations(){

        return this.localDeclarations;
    }

    /** Add a <code>FilterCondition</code> to this rule.
     *
     *  @param filterCondition The <code>FilterCondition</code> to add.
     */
    public void addFilterCondition(FilterCondition filterCondition){

	this.decl = false;

        if ( this.filterConditions == Collections.EMPTY_SET )

            this.filterConditions = new HashSet();

        this.filterConditions.add( filterCondition );
    }

    /** Add a consistent <code>AssignmentCondition</code> to this rule.
     *
     *  @param assignmentCondition the <code>AssignmentCondition</code> to add.
     */
    public void addAssignmentCondition(AssignmentCondition assignmentCondition){

	this.decl = false;

        if ( this.assignmentConditions == Collections.EMPTY_SET )

            this.assignmentConditions = new HashSet();


        this.assignmentConditions.add( assignmentCondition );

        Declaration decl = assignmentCondition.getTargetDeclaration();

        if ( ! this.parameterDeclarations.contains( decl ) )

            addLocalDeclaration( decl );
    }

    /** Retrieve the <code>Set</code> of <code>FilterConditions</code> for
     *  this rule.
     *
     *  @return The <code>Set</code> of <code>FilterConditions</code>.
     */
    public Set getFilterConditions(){

        return this.filterConditions;
    }

    /** Retrieve the <code>Set</code> of <code>AssignmentConditions</code> for
     *  this rule.
     *
     *  @return The <code>Set</code> of <code>AssignmentConditions</code>.
     */
    public Set getAssignmentConditions(){

        return this.assignmentConditions;
    }

    /** Set the <code>Action</code> that is associated with the
     *  successful match of this rule.
     *
     *  @param action The <code>Action</code> to attach to this <code>Rule</code>.
     */
    public void setAction(Action action){

	if (!this.name.equals(action.getRuleName()))

	    throw new IllegalStateException("BBBUGGGG! Action adding to wrong rule.");
	else
	    this.action = action;
    }

    /** Retrieve the <code>Action</code> associated with this <code>Rule</code>.
     *
     *  @return The <code>Action</code>.
     */
    public Action getAction(){

        return this.action;
    }
}
